# -*- coding: utf-8 -*-
"""
   Wiki+

   Copyright (C) 2005 Yong-Jin Cho <yongjin.cho@gmail.com>
   Wiki+ is distributed under GNU GPL. See COPYING for details.

   This program is based on PikiPiki which is a simple wiki engine.
   PikiPiki project homepage is http://sourcefrog.net/projects/piki.

   For more information, visit http://wikiplus.kldp.net.
"""
# Imports -----------------------------------------------------------
import cgi, sys, string, os, fcntl, re, errno, time, urllib
import config



# Exceptions --------------------------------------------------------

# for page class
PageError = 'PageError'
NoPage = 'NoPage'
EmptyText = 'EmptyText'
UnchangedText = 'UnchangedText'
UpdateConflict = 'UpdateConflict'

# for action handler
UnknownAction = 'UnknownAction'



# Page class --------------------------------------------------------

class Page:
    """Page management class

       Directory Structure
       -------------------
        data/ --+-- text/ ---+-- pageid1
                |            `-- pageid2
                |
                +-- backup/ -+-- pageid1            --> editlog
                |            +-- pageid1.timestamp
                |            +-- pageid1.timestamp
                |            +-- pageid2
                |            `-- pageid2.timestamp
                |
                `-- editlog

       Page ID
       -------
        quoted page name

       Editlog File Structure
       ----------------------
        +-------------------------------------------------------+
        |page id<TAB>modified time<TAB>host<NEWLINE>|
        |    .             .             .         .            |
        |                                                       |
    """

    def __init__(self, name, timestamp=None):
        self.name = name
        self.url_name = urlencode(name)
        self.file_name = self.url_name.replace('%', '_')
        self.timestamp = timestamp


    def _text_path(self):
        if not self.timestamp:
            return os.path.join(config.text_dir, self.file_name)
        else:
            return os.path.join(config.backup_dir,
                                        self.file_name + '.' + self.timestamp)

    def _log_path(self):
        return os.path.join(config.backup_dir, self.file_name)

    def _tmp_path(self):
        return os.path.join(config.data_dir,
                        ('#' + self.file_name + '.' + str(os.getpid()) + '#'))

    def is_current(self):
        if not self.timestamp:
            return True
        else:
            return False


    def exists(self):
        return os.path.exists(self._text_path())


    def link(self):
        if not self.is_current():
            raise PageError
        if self.exists():
            return link_tag(self.url_name, self.name)
        else:
            return link_tag(self.url_name, '?') + self.name


    def get_mtime(self):
        try:
            return os.path.getmtime(self._text_path())
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er


    def get_size(self):
        try:
            return os.path.getsize(self._text_path())
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er


    def get_text(self):
        try:
            text_file = open(self._text_path(), 'rb')
        except IOError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er
        try:
            text = text_file.read()
        finally:
            text_file.close()
        return text


    def update_text(self, timestamp, newtext):
        if not self.is_current():
            raise PageError
        newtext = newtext.replace('\r', '')
        if not newtext:
            raise EmptyText

        tmp_path = self._tmp_path()
        tmp_file = open(tmp_path, 'wb')
        tmp_file.write(newtext)
        tmp_file.close()

        text_path = self._text_path()
        if self.exists():
            mtime = self.get_mtime()
            if timestamp != str(mtime):
                raise UpdateConflict
            elif newtext == self.get_text():
                raise UnchangedText
            os.rename(text_path, os.path.join(config.backup_dir,
                                            self.file_name + '.' + str(mtime)))

        os.chmod(tmp_path, 0666)
        os.rename(tmp_path, text_path)

        timestamp = str(self.get_mtime())
        host = os.environ.get('REMOTE_ADDR', '')
        editlog_add(self._log_path(), self.file_name, timestamp, host)
        os.chmod(self._log_path(), 0666)
        editlog_add(config.editlog_path,
                                      self.file_name, timestamp, host)


    def get_backup_list(self):
        tmp_list = os.listdir(config.backup_dir)
        prefix = self.file_name + '.'
        tmp_list = filter(lambda x: x.startswith(prefix), tmp_list)
        return map(lambda x: x.split('.')[1], tmp_list)


    def send_page(self, msg=None):
        link = '%s?action=search&amp;method=full&amp;value=%s' % \
                                                (self.url_name, self.url_name)
        send_title(self.name, link, msg)
 
        try:
            text = self.get_text()
        except NoPage:
            print link_tag(self.url_name + '?action=edit', 'Create this page')
            last_modified = None
        else:
            Formatter().format(text)
            mtime = self.get_mtime()
            mod_str = time.strftime(config.datetime_fmt,
                                            time.localtime(self.get_mtime()))
            last_modified = link_tag(self.url_name + '?action=info', mod_str)
 
        send_footer(self.name, True, last_modified)


    def send_editor(self, preview=False, **kw):
        if not self.is_current():
            raise PageError
        if not preview:
            send_title('Edit "%s"' % self.name)
            if self.exists():
                timestamp = str(self.get_mtime())
                newtext = self.get_text()
            else:
                timestamp = ''
                newtext = 'Describe %s here.' % self.name
        else:
            send_title('Preview of "%s"' % self.name)
            timestamp = kw['timestamp']
            newtext = kw['newtext'].replace('\r', '')

        print '<form method="post" action="%s/%s#preview">' % \
              (get_scriptname(), self.url_name)
        print '<input type="hidden" name="action" value="update" />'
        print '<input type="hidden" name="timestamp" value="%s" />' % timestamp
        print '<textarea name="newtext" rows="18" cols="90" ' \
              'style="width:95%%;word-wrap:virtual;">%s</textarea><br />' % newtext
        print '<input type="submit" name="save" value="Save" /> ' \
              '<input type="submit" name="preview" value="Preview" /></form>'

        if preview:
            print '<a name="preview"></a><hr />'
            Formatter().format(newtext)

        send_footer(self.name, False)


def get_page_list():
    tmp_list = map(lambda x:x.replace('_', '%'), os.listdir(config.text_dir))
    return map(urldecode, tmp_list)



# Editlog -----------------------------------------------------------

# Functions to keep track of when people have changed pages, so we can
# do the recent changes page and so on.
# The editlog is stored with one record per line, as tab-separated
# words: pageid, timestamp, host, comment

def editlog_add(log_path, pageid, timestamp, host):
    entry = string.join((pageid, timestamp, host), "\t") + "\n"
    editlog = open(log_path, 'ab')
    try: 
        fcntl.flock(editlog.fileno(), fcntl.LOCK_EX)
        editlog.seek(0, 2)                  # to end
        editlog.write(entry)
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()


def editlog_raw_lines(log_path):
    editlog = open(log_path, 'rb')
    try:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_SH)
        return editlog.readlines()
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()



# Formatter ---------------------------------------------------------

class Formatter:
    def __init__(self):
        self._block_re = re.compile(r'^(?:' \
            r'(?P<heading>(?P<hmarker>={1,6})\s.*\s(?P=hmarker))|' \
            r'(?P<rule>-{4,})|' \
            r'(?P<list>(?P<indent>\s+)(?:(?P<ol>\d+\.\s)|(?P<ul>\*\s)).*)|' \
            r'(?P<paragraph>.+)|' \
            r'(?P<blank>)' \
            r')$')
        self._cur_block = 'blank'
        self._list_stack = [(0, None)]      # (level, order)

        self._inline_re = re.compile(
            r"(?P<emph>'{2,3})|" \
            r'(?P<word>\b(?:[A-Z][a-z]+){2,}\b)|' \
            r'(?P<eword>\[[^\[\]]+\])|' \
            r'(?P<url>(?:http|ftp|nntp|news|mailto)\:[^\s\'"]+\S)|' \
            r'(?P<email>[-\w._+]+\@[\w.-]+)|' \
            r'(?P<macro>\[\[[A-Za-z][A-Za-z0-9]*(?:\(.*\))?\]\])|' \
            r'(?P<code>\{\{\{[^(?:\}\}\})]+\}\}\})|' \
            r'(?P<pre>(?:\{\{\{|\}\}\}))|' \
            r'(?P<ent>[<>&])')
        self._in_pre = False
        self._is_em = self._is_b = False


    def format(self, raw):
        raw_lines = raw.expandtabs().split('\n')
        for raw_line in raw_lines:
            if not self._in_pre:
                match = self._block_re.match(raw_line)
                for block, s in match.groupdict().items():
                    if s != None and hasattr(self, '_' + block + '_block'):
                        self._change_block(block)
                        apply(getattr(self, '_' + block + '_block'), (match,))
                        break
            else:
                print self._inline(raw_line)

        self._change_block('blank')


    # block formatting part
    def _change_block(self, new_block):
        old_block = self._cur_block
        self._cur_block = new_block
        if new_block != old_block:
            if old_block == 'list':
                while self._list_stack[-1][0]:
                    prev_order = self._list_stack.pop()[1]
                    print '</li></%s>' % prev_order
            elif old_block == 'paragraph':
                print '</p>'

            if new_block == 'paragraph':
                print '<p>'


    def _heading_block(self, match):
        level = len(match.group('hmarker'))
        print '<h%d>' % level
        print self._inline(match.group(0)[level + 1: -(level + 1)])
        print '</h%d>' % level

    def _rule_block(self, match):
        if len(match.group(0)) <= 4:
            print '<hr />'
        else:
            print '<hr size="%d" />' % (len(match.group(0)) - 2)

    def _list_block(self, match):
        if match.group('ul'):
            order = 'ul'
        else:
            order = 'ol'
        level = len(match.group('indent'))
        prev_level = self._list_stack[-1][0]

        if level == prev_level:
            print '</li><li>'
        elif level > prev_level:
            print '<%s><li>' % order
            self._list_stack.append((level, order))
        else:
            while level < prev_level:
                prev_order = self._list_stack.pop()[1]
                print '</li></%s>' % prev_order
                prev_level = self._list_stack[-1][0]
            print '</li><li>'

        print self._inline(match.group(0)[level + len(match.group(order)):])

    def _paragraph_block(self, match):
        print self._inline(match.group(0))
        if config.auto_linebreak and not self._in_pre:
            print '<br />'

    def _blank_block(self, match):
        pass


    # inline replacement part
    def _inline(self, line):
        return self._inline_re.sub(self._replace, line)


    def _replace(self, match):
        for type, hit in match.groupdict().items():
            if hit:
                return apply(getattr(self, '_' + type + '_repl'), (hit,))


    def _emph_repl(self, word):
        if not self._in_pre:
            if len(word) == 3:
                self._is_b = not self._is_b
                return ['</b>', '<b>'][self._is_b]
            else:
                self._is_em = not self._is_em
                return ['</em>', '<em>'][self._is_em]
        else:
            return word

    def _word_repl(self, word):
        if not self._in_pre:
            return Page(word).link()
        else:
            return word

    def _eword_repl(self, word):
        if not self._in_pre:
            return Page(word[1:-1]).link()
        else:
            return word

    def _url_repl(self, word):
        if not self._in_pre:
            ext = word[word.rfind('.') + 1:]
            if ext in ('gif', 'jpg', 'bmp', 'png'):
                return '<img src="%s" alt="%s" />' % (word, word)
            else:
                return '<a href="%s">%s</a>' % (word, word)
        else:
            return word

    def _email_repl(self, word):
        if not self._in_pre:
            return '<a href="mailto:%s">%s</a>' % (word, word)
        else:
            return word

    def _macro_repl(self, word):
        if not self._in_pre:
            pos = word.find('(')
            if pos == -1:
                name = word[2:-2]
                arg = ''
            else:
                name = word[2:pos]
                arg = word[pos+1:-3]
            return expand_macro(name, arg)
        else:
            return word

    def _code_repl(self, word):
        if not self._in_pre:
            return '<code>%s</code>' % word[3:-3]
        else:
            return word

    def _pre_repl(self, word):
        if not self._in_pre and word == '{{{':
            self._in_pre = True
            return '<pre>'
        elif self._in_pre and word == '}}}':
            self._in_pre = False
            return '</pre>'
        else:
            return word

    def _ent_repl(self, s):
        return {'&': '&amp;', '<': '&lt;', '>': '&gt;'}[s]



# Macro functions ---------------------------------------------------

def init_macro(page_name, form):
    global _macro
    _macro = Macro(page_name, form)


def expand_macro(name, arg):
    global _macro
    return _macro.expand(name, arg)


class Macro:
    def __init__(self, page_name, form):
        self.page_name = page_name
        self.form = form


    def expand(self, name, arg):
        if hasattr(self, '_macro_' + name):
            return apply(getattr(self, '_macro_' + name), (arg,))
        else:
            return '<i>"%s" macro does not exist.</i>' % name


    def _macro_RecentChanges(self, arg):
        import cStringIO

        lines = editlog_raw_lines(config.editlog_path)
        lines.reverse()
    
        buf = cStringIO.StringIO()
        ratchet_day = None
        done_names = {}
        for line in lines:
            page_name, mtime, host = line[:-1].split('\t')
            page_name = unquote_name(page_name)

            if done_names.has_key(page_name):
                continue
            done_names[page_name] = 1

            # year, month, day, DoW
            time_tuple = time.localtime(float(mtime))
            day = tuple(time_tuple[0:3])
            if day != ratchet_day:
                if ratchet_day:
                    buf.write('</ul>')
                buf.write('<h3>%s</h3><ul>' % time.strftime(config.date_fmt, time_tuple))
                ratchet_day = day

            buf.write('<li>')
            buf.write(Page(page_name).link())
            buf.write(' . . . . ')
            if config.show_hosts:
                try:
                    buf.write(socket.gethostbyaddr(host)[0])
                except:
                    buf.write(host)
                buf.write(' ')

            if config.changed_time_fmt:
                buf.write(time.strftime(config.changed_time_fmt, time_tuple))

            buf.write('</li>')
        buf.write('</ul>')

        return buf.getvalue()

    
    def _macro_FullSearch(self, arg):
        url_name = urlencode(self.page_name)
        return '<form action="%s/%s" method="get">' \
               '<input type="hidden" name="action" value="search" />' \
               '<input type="hidden" name="method" value="full" />' \
               '<input type="text" name="value" size="30" /> ' \
               '<input type="submit" value="Search" />' \
               '</form>' % (get_scriptname(), url_name)


    def _macro_TitleSearch(self, arg):
        url_name = urlencode(self.page_name)
        return '<form action="%s/%s" method="get">' \
               '<input type="hidden" name="action" value="search" />' \
               '<input type="hidden" name="method" value="title" />' \
               '<input type="text" name="value" size="30" /> ' \
               '<input type="submit" value="Search" />' \
               '</form>' % (get_scriptname(), url_name)


    def _macro_TitleIndex(self, arg):
        import cStringIO

        page_list = get_page_list()
        tmp_list = map(lambda x: x.decode(config.encoding), page_list)
        tmp_list = map(lambda x: (x.upper(), x), tmp_list)
        tmp_list.sort()
        page_list = map(lambda x: x[1], tmp_list)

        buf = cStringIO.StringIO()
        index_letters = []
        current_letter = None
        for name in page_list:
            letter = get_index_group(name)
            letter = letter.upper()
            if letter not in index_letters:
                index_letters.append(letter)

            if letter != current_letter:
                if current_letter:
                    buf.write('</ul>')
                encoded_letter = letter.encode(config.encoding)
                buf.write('<h2><a name="%s" href="#top">%s</a></h2><ul>' %
                                                (encoded_letter, encoded_letter))
                current_letter = letter

            buf.write('<li>%s</li>' % Page(name.encode(config.encoding)).link())
        buf.write('</ul>')

        index = map(lambda x: x.encode(config.encoding), index_letters)
        index = map(lambda x: '<a href="#%s">%s</a> ' % (x, x), index)
        top = '<a name="top"></a>'

        return top + ' | '.join(index) + buf.getvalue()


    def _macro_PageCount(self, arg):
        page_list = get_page_list()
        return str(len(page_list))


    def _macro_DueDate(self, arg):
        import math
        try:
            t = time.strptime(arg, '%Y%m%d')
        except ValueError, er:
            return '<i>%s</i>' % er

        secs = time.mktime(t)
        cur_secs = time.time()
        dist_day = int(math.ceil((secs - cur_secs) / (24 * 60 * 60)))

        return str(abs(dist_day))


    def _macro_UploadFile(self, arg):
        return '<form action="%s/%s" method="post" enctype="multipart/form-data">' \
               '<input type="hidden" name="action" value="upload" />' \
               '<input type="file" name="file" size="40" /><br />' \
               'Rename (optional): <input type="text" name="rename" size="30" />' \
               '<br /><input type="submit" value="Upload" />' \
               '</form>' % (get_scriptname(), self.page_name)


    def _macro_UploadedFile(self, arg):
        file_name = arg
        img_exts = ('.jpg', '.gif', '.bmp', '.png', 'jpeg')
        name = cgi.escape(file_name)

        if os.path.exists(os.path.join(config.upload_dir, file_name)):
            url = config.upload_url + '/' + urlencode(name)

            if os.path.splitext(file_name)[1] in img_exts:
                return '<img src="%s" alt="%s" />' % (url, name)
            else:
                return '<a href="%s">%s</a>' % (url, name)
        else:
            return '<i>"%s" file does not exist.</i>' % name


    def _macro_UploadedFiles(self, arg):
        import cStringIO

        buf = cStringIO.StringIO()

        file_list = os.listdir(config.upload_dir)
        buf.write('<table><tr align="center">')
        buf.write('<th>Name</th><th>Size</th><th>Date</th></tr>')
        for file_name in file_list:
            file_path = os.path.join(config.upload_dir, file_name)

            name = cgi.escape(file_name)
            url = config.upload_url + '/' + urlencode(file_name)
            size = os.path.getsize(file_path)
            mtime = os.path.getmtime(file_path)

            buf.write('<tr><td><a href="%s">%s</a></td>' % (url, name))
            buf.write('<td align="right">%s</td>' % str_size(size))
            buf.write('<td>%s</td></tr>' % \
                      time.strftime(config.datetime_fmt, time.localtime(mtime)))
        buf.write('</table>')
        buf.write('<b>Total %d files are uploaded.</b>' % len(file_list))

        return buf.getvalue()


    def _macro_ISBN(self, arg):
        isbn = arg
        amazon_link = 'http://www.amazon.com/exec/obidos/ISBN=%s'
        amazon_img = 'http://images.amazon.com/images/P/%s.01.MZZZZZZZ.gif'
        aladdin_link = 'http://www.aladdin.co.kr/catalog/book.asp?ISBN=%s'
        aladdin_img = 'http://www.aladdin.co.kr/Cover/%s_1.gif'

        if isbn[:2] == '89':
            link, img = aladdin_link % isbn, aladdin_img % isbn
        else:
            link, img = amazon_link % isbn, amazon_img % isbn
    
        return '<a href="%s"><img src="%s" alt="%s" /></a>' % (link, img, isbn) 



# Utility functions -------------------------------------------------

def emit_header():
    print "Content-type: text/html"
    print


def urlencode(s):
    t = urllib.quote(s, '')
    t = t.replace('-', '%2D')
    t = t.replace('.', '%2E')
    return t.replace('_', '%5F')


def urldecode(s):
    return urllib.unquote(s)


def quote_name(name):
    return urlencode(name).replace('%', '_')


def unquote_name(name):
    return urldecode(name.replace('_', '%'))


def quote_path(path):
    path = path.replace('/', '_')
    path = path.replace('\\', '_')
    return path.replace(':', '_')


def get_scriptname():
    return os.environ.get('SCRIPT_NAME', '')


def get_index_group(name):
    c = name[0]
    if u'\uAC00' <= c <= u'\uD7AF': # Hangul Syllables
        return unichr(0xac00 + (int(ord(c) - 0xac00) / 588) * 588)
    else:
        return c


def get_basename(path):
    import posixpath, dospath, macpath
    for pathmodule in (posixpath, dospath, macpath):
        basename = pathmodule.basename(path)
        if basename != path:
            return basename
    return path


def str_size(size):
    if size < 1024:
        return str(size)
    elif size < 1024 ** 2:
        ksize = size / 1024.0
        if int(ksize) < 10:
            return str(round(ksize, 1)) + 'K'
        return str(int(ksize)) + 'K'
    else:
        msize = size / (1024.0 ** 2)
        if int(msize) < 10:
            return str(round(msize, 1)) + 'M'
        return str(int(msize)) + 'M'


def link_tag(link, text=None):
    if not text:
        text = link
    return '<a href="%s/%s">%s</a>' % (get_scriptname(), link, text)


def send_title(title, link=None, msg=None):
    splited_title = re.sub(r'([a-z])([A-Z])', r'\1 \2', title)

    print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" ' \
          '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
    print '<html><head><title>%s</title>' % splited_title
    print '<meta http-equiv="content-type" content="text/html;' \
          'charset=%s" />' % config.encoding
    print '<link rel="stylesheet" type="text/css" href="%s" /></head>' \
          '<body>' % config.css_url
    print link_tag(urlencode(config.front_page),
                  '<img class="logo" src="%s" alt="logo" />' % config.logo_url)
    # goto form
    print '<div class="goto">' \
          '<form action="%s" method="get">' % get_scriptname()
    print 'GO <input type="hidden" name="action" value="goto" />' \
          '<input type="text" name="value" size="10" /></form></div>' \
          '<div class="title">'
    if link:
        print link_tag(link, splited_title)
    else:
        print title
    print '</div>'
    print link_tag(urlencode(config.recent_changes), config.recent_changes)
    print ' | '
    print link_tag(urlencode(config.find_page), config.find_page)
    print ' | '
    print link_tag(urlencode(config.title_index), config.title_index)
    print '<hr /><div class="body">'
    if msg:
        print '%s<hr />' % msg


def send_footer(page_name, editable=True, mod_string=None):
    print '</div><hr /><div class="footmenu">'
    if editable:
        print link_tag(urlencode(page_name) + '?action=edit', config.edit_page)
    else:
        print link_tag(urlencode(page_name), config.show_page)
    if mod_string:
        print ' (last modified %s)' % mod_string
    print '</div><div class="power">Powered by ' \
          '<a href="http://wikiplus.kldp.net">Wiki+</a><br />' \
          '<a href="http://validator.w3.org/check/referer">XHTML 1.0</a> | ' \
          '<a href="http://jigsaw.w3.org/css-validator/check/referer">' \
          'CSS 2.0</a></div></body></html>'



# Action handlers ---------------------------------------------------

def do_show(page_name, form):
    pg = Page(page_name, form.getvalue('timestamp'))
    if pg.is_current():
        msg = None
    else:
        try:
            mtime = pg.get_mtime()
        except NoPage:
            msg = '<b>No revision exists.</b>'
        else:
            date = time.strftime(config.datetime_fmt, time.localtime(mtime))
            msg = '<b>Revision as of %s</b>' % date
    pg.send_page(msg)


def do_edit(page_name, form):
    Page(page_name).send_editor()


def do_update(page_name, form):
    timestamp = form.getvalue('timestamp')
    newtext = form.getvalue('newtext', '')

    pg = Page(page_name)
    if form.getvalue('save'):
        try:
            pg.update_text(timestamp, newtext)
        except EmptyText:
            msg = '<b>You cannot save empty pages.</b>'
        except UnchangedText:
            msg = '<b>You cannot save unchanged pages.</b>'
        except UpdateConflict:
            msg = '<p><b>Sorry, someone else saved the page while you edited' \
                  ' it.</b></p><p>Please do the following: Cut&amp;paste you' \
                  'r changes from the following box. Then click EditPage aga' \
                  'in(or shift-click it to open in a new window). Now re-add' \
                  ' your changes to the current page contents.</p><p><em>Do ' \
                  'not just replace the content editbox with your version of' \
                  ' the page, because that would delete the changes of the o' \
                  'ther person, which is excessively rude!</em></p><p>Follow' \
                  'ing is the page you edited.</p>'
            msg += '<textarea wrap="virtual" rows="17" cols="80">%s' \
                   '</textarea>' % newtext
        else:
            msg = '<b>Thankyou for your changes. Your attention to detail is' \
                  ' appreciated.</b>'
        pg.send_page(msg)

    else:   # if preview
        pg.send_editor(True, timestamp=timestamp, newtext=newtext)


def do_text(page_name, form):
    pg = Page(page_name, form.getvalue('timestamp'))
    try:
        text = pg.get_text()
    except NoPage:
        pg.seng_page('<b>No revision exists.</b>')
    else:
        print '<pre>'
        print text
        print '</pre>'


def do_info(page_name, form):
    def print_info_row(pg):
        mtime = pg.get_mtime()
        print '<tr><td>%s</td>' % \
                    time.strftime(config.datetime_fmt, time.localtime(mtime))
        print '<td align="right">%s</td><td>' % str_size(pg.get_size())
        if pg.is_current():
            url_suffix = ''
        else:
            url_suffix = '&timestamp=' + pg.timestamp
        print link_tag(pg.url_name + '?action=show' + url_suffix, 'show')
        print link_tag(pg.url_name + '?action=text' + url_suffix, 'text')
        print '</td></tr>'

    pg = Page(page_name)
    backup_list = pg.get_backup_list()
    backup_list.sort()
    backup_list.reverse()

    if len(backup_list):
        send_title('Info of "' + page_name + '"')
        print '<table><tr><th>Date</th><th>Size</th><th>Actions</th></tr>'
        if pg.exists():
            print_info_row(pg)
        for timestamp in backup_list:
            pg = Page(page_name, timestamp)
            print_info_row(pg)
        print '</table>'
        send_footer(page_name, False)
    else:
        pg.send_page('<b>No older revision exists.</b>')


def do_goto(page_name, form):
    page_name = form.getvalue('value')
    init_macro(page_name, form)         # reset global variables for macro
    Page(page_name).send_page()


def do_search(page_name, form):
    method = form.getvalue('method')
    if method not in ('title', 'full'):
        method = 'title'

    needle = form.getvalue('value', '')
    if method == 'title':
        send_title("Title search for \"" + needle + '"')
    else:
        send_title('Full search for "%s"' % (needle))

    needle_re = re.compile(needle, re.IGNORECASE)
    page_list = get_page_list()
    if method == 'title':
        hits = filter(needle_re.search, page_list)
    else:
        hits = []
        for page in page_list:
            body = Page(page).get_text()
            match = needle_re.findall(body)
            count = len(needle_re.findall(body))
            if count:
                hits.append((count, page))

        # The default comparison for tuples compares elements in order,
        # so this sorts by number of hits
        hits.sort()
        hits.reverse()

    print '<ul>'
    if method == 'title':
        for page in hits:
            print '<li>' + Page(page).link() + '</li>'
    else:
        for (count, page) in hits:
            print '<li>' + Page(page).link()
            print ' . . . . ' + `count`
            print ['match', 'matches'][count != 1] + '</li>'
    print '</ul>'

    print "<p>%d hits " % len(hits)
    print " out of %d pages searched." % len(page_list)

    send_footer(page_name, 1)


def do_upload(page_name, form):
    import shutil

    if not form.getvalue('file') or not form['file'].file:
        msg = '<b>File was not trasferred.</b>'
    else:
        base_name = form.getvalue('rename')
        if not base_name:
            base_name = get_basename(form['file'].filename)

        name, ext = os.path.splitext(quote_path(base_name))
        if ext in config.allowed_ext:
            # conflict resolution
            suffix = ''
            while os.path.exists(os.path.join(config.upload_dir,
                                              name + suffix + ext)):
                if not suffix:
                    suffix = '0'
                else:
                    suffix = str(int(suffix) + 1)
            file_name = name + suffix + ext            
            file_path = os.path.join(config.upload_dir, file_name)

            # copy
            dest_file = open(file_path, 'wb')
            shutil.copyfileobj(form['file'].file, dest_file, 1024)
            dest_file.close()
            os.chmod(file_path, 0666)

            msg = '<b>"%s" was successfully uploaded.</b>' % file_name
        else:
            msg = '<b>"%s" is not allowed extension.</b>' % ext
    
    Page(page_name).send_page(msg)


def get_handler(action):
    handlers = {
        'show':     do_show,
        'edit':     do_edit,
        'update':   do_update,
        'text':     do_text,
        'info':     do_info,
        'goto':     do_goto,
        'search':   do_search,
        'upload':   do_upload
    }
    if not handlers.has_key(action):
        raise UnknownAction, action
    return handlers[action]



# CGI main ----------------------------------------------------------

def main():
    try:
        emit_header()

        # get request
        path_info = os.environ.get('PATH_INFO', '')
        if len(path_info) > 1:
            page_name = path_info[1:]
        else:
            page_name = config.front_page
        form = cgi.FieldStorage()
        action = form.getvalue('action', 'show')

        # inititalize macro
        init_macro(page_name, form)

        # handle request
        handler = get_handler(action)
        handler(page_name, form)

    except UnknownAction, action:
        print 'Unknown action: ' + action

    except:
        cgi.print_exception()

    sys.stdout.flush()
